// LinkerNode.cpp
//------------------------------------------------------------------------------

// Includes
//------------------------------------------------------------------------------
#include "LinkerNode.h"

#include <Tools/FBuild/FBuildCore/FBuild.h>
#include <Tools/FBuild/FBuildCore/Flog.h>
#include <Tools/FBuild/FBuildCore/Graph/DLLNode.h>
#include <Tools/FBuild/FBuildCore/Graph/FileNode.h>
#include <Tools/FBuild/FBuildCore/Graph/LibraryNode.h>
#include <Tools/FBuild/FBuildCore/Graph/ObjectNode.h>
#include <Tools/FBuild/FBuildCore/Graph/NodeGraph.h>

#include <Core/FileIO/FileIO.h>
#include <Core/FileIO/FileStream.h>
#include <Core/Process/Process.h>
#include <Core/Strings/AStackString.h>

// CONSTRUCTOR
//------------------------------------------------------------------------------
LinkerNode::LinkerNode( const AString & linkerOutputName,
						 const Array< Node * > & inputLibraries,
						 const AString & linker,
						 const AString & linkerArgs,
						 uint32_t flags )
: FileNode( linkerOutputName, Node::FLAG_NONE )
, m_StaticDependencies( inputLibraries.GetSize(), false )
, m_Flags( flags )
{
	m_LastBuildTimeMs = 20000;

	// depend on everything we'll link together
	ASSERT( inputLibraries.IsEmpty() == false );
	m_StaticDependencies = inputLibraries;

	// store options we'll need to use dynamically
	m_Linker = linker; // TODO:C This should be a node
	m_LinkerArgs = linkerArgs;
}

// DESTRUCTOR
//------------------------------------------------------------------------------
LinkerNode::~LinkerNode()
{
}

// DoBuild
//------------------------------------------------------------------------------
/*virtual*/ Node::BuildResult LinkerNode::DoBuild( Job * UNUSED( job ) )
{
	// need building...
	FLOG_BUILD( "%s: %s\n", GetDLLOrExe(), GetName().Get() );

	// Format compiler args string
	AStackString< 4 * KILOBYTE > fullArgs;
	fullArgs = m_LinkerArgs;

	// substitute "function variables" if needed
	AStackString< 4 * KILOBYTE > inputFiles;
	GetInputFiles( inputFiles );
	fullArgs.Replace( "%1", inputFiles.Get() );
	fullArgs.Replace( "%2", m_Name.Get() );

	// use the exe launch dir as the working dir
	const char * workingDir = nullptr;

	const char * environment = FBuild::Get().GetEnvironmentString();

	FLOG_INFO( "%s %s\n", m_Linker.Get(), fullArgs.Get() );

	// spawn the process
	Process p;
	bool spawnOK = p.Spawn( m_Linker.Get(),
							fullArgs.Get(),
							workingDir,
							environment );

	if ( !spawnOK )
	{
		FLOG_ERROR( "Failed to spawn process for %s creation for '%s'", GetDLLOrExe(), GetName().Get() );
		return NODE_RESULT_FAILED;
	}

	// capture all of the stdout and stderr
	AutoPtr< char > memOut;
	AutoPtr< char > memErr;
	uint32_t memOutSize = 0;
	uint32_t memErrSize = 0;
	p.ReadAllData( memOut, &memOutSize, memErr, &memErrSize );

	ASSERT( !p.IsRunning() );
	// Get result
	int result = p.WaitForExit();

	if ( result != 0 )
	{
		if ( memOut.Get() ) { FLOG_ERROR_DIRECT( memOut.Get() ); }
		if ( memErr.Get() ) { FLOG_ERROR_DIRECT( memErr.Get() ); }
	}

	// did the executable fail?
	if ( result != 0 )
	{
		FLOG_ERROR( "Failed to build %s (error %i) '%s'", GetDLLOrExe(), result, GetName().Get() );
		return NODE_RESULT_FAILED;
	}

	// record time stamp for next time
	m_Stamp = FileIO::GetFileLastWriteTime( m_Name );
	ASSERT( m_Stamp );

	return NODE_RESULT_OK;
}

// GetInputFiles
//------------------------------------------------------------------------------
void LinkerNode::GetInputFiles( AString & inputFiles ) const
{
	for ( Array< Node * >::Iter i = m_StaticDependencies.Begin();
		  i != m_StaticDependencies.End();
		  i++ )
	{
		Node * n( *i );

		if ( n->GetType() == Node::LIBRARY_NODE )
		{
			if ( GetFlag( LINK_FLAG_MSVC ) && GetFlag( LINK_FLAG_DLL ) )
			{
				// If we're building a DLL with MSVC, we must link the obj files directly
				// because the lib doesn't contain the export information
				const Array< Node * > & libNodes = ( *i )->GetDynamicDependencies();
				const Array< Node * >::Iter end = libNodes.End();
				for ( Array< Node * >::Iter it = libNodes.Begin();
					  it != end; 
					  ++it )
				{
					Node * subNode = *it;
					if ( subNode->GetType() == Node::OBJECT_NODE )
					{
						// handle pch files - get path to matching object
						const ObjectNode * on = subNode->CastTo< ObjectNode >();
						if ( on->IsCreatingPCH() )
						{
							inputFiles += subNode->GetName();
							inputFiles += ".obj ";
							continue;
						}

						inputFiles += subNode->GetName();
						inputFiles += ' ';
					}
					else
					{
						ASSERT( false && "Unexpected node type" );
					}
				}
			}
			else
			{
				// not building a DLL, so link the lib directly
				inputFiles += ( *i )->GetName();
			}
		}
		else if ( n->GetType() == Node::DLL_NODE )
		{
			// for a DLL, link to the import library
			DLLNode * dllNode = n->CastTo< DLLNode >();
			AStackString<> importLibName;
			dllNode->GetImportLibName( importLibName );
			inputFiles += importLibName;
		}
		else
		{
			// link anything else directly
			inputFiles += ( *i )->GetName();
		}

		inputFiles += ' ';
	}
}

// GetDependencies
//------------------------------------------------------------------------------
/*virtual*/ const Array< Node * > & LinkerNode::GetStaticDependencies() const
{
	return m_StaticDependencies;
}

// Save
//------------------------------------------------------------------------------
/*virtual*/ bool LinkerNode::Save( IOStream & stream, bool remote ) const
{
	NODE_SAVE( m_Name );
	NODE_SAVE( m_Linker );
	NODE_SAVE( m_LinkerArgs );
	NODE_SAVE_DEPS( m_StaticDependencies );
	NODE_SAVE( m_Flags );
	return true;
}

//------------------------------------------------------------------------------
